#include "raw.h"

/* 
   reads in raw ABI files produced by AGRF (-a command line switch)
   Originally written by M. Bahlo, last mod to orginal version 18/4/2001 
   
   $Id: rawabi.c,v 2.7 2002/07/27 09:24:53 wilkinson Exp $
   $Log: rawabi.c,v $
   Revision 2.7  2002/07/27 09:24:53  wilkinson
   made reading of pedigree files more robust.
   The program automatically recognises if there is a header line present
   Note if a header line is present, it must consist of single words (no spaces)
   not containing numbers

   Revision 2.6  2002/07/18 01:33:39  wilkinson
   modified reading of marker and distance (recomb frac) files to ignore
   lines beginning with a '#'. When reading distances, it will store the
   distance after the # and add that to the the distance to the next marker.

   Revision 2.5  2001/11/16 00:27:16  wilkinson
   Enabled the use of an input file in placed of command line arguments.
   The file template file inputs lists the general style required.

   Revision 2.4  2001/11/14 04:33:45  wilkinson
   add warnings if unable to open files for reading or writing

   Revision 2.3  2001/11/09 01:30:23  wilkinson
   added a debug flag on the command line
   added a check to ensure that zeroed frequencies (ie present in affected or
   sibs but not in founders) are set to a value of MINFREQUENCY

   Revision 2.2  2001/10/31 03:02:06  wilkinson
   Bug detected in writing of pre files.
   If there is no data on a marker specified in the marker list, the program
   did not detect this. When writing the .pre file, the marker was expected
   to be present, and an array was accessed out of bounds in this case
   (resulting in a random number of expected alleles).
   The case of a no data on a given marker is now detected, and a warning
   message is issued to the user followed by program termination

   Revision 2.1  2001/10/11 06:47:16  wilkinson
   code modified to save raw allele values as well.  These will then be
   written out to a seperate .pre file

   Revision 2.0  2001/10/08 02:31:32  wilkinson
   Fixed bug with reading in of marker files (reading in one less than
   the desired number).
   Made getline function more robust to allow multiple spaces or tabs to
   be handled (originally all field could only be seperated by a single space
   or tab).  getline also now handles whitespace after the field being read
   Program also handles extra blank lines at the end of input files


*/

char mark_ord[MAXLOCI][MAXCHAR];
char sample_id[MAXIND][MAXCHAR];
DATA *dat;


/* Get memory for DATA, use known number of Individuals and Loci */
DATA *allocate_data(int NoInd, int NoLoci) {
  int i,j;

  dat               = (DATA*)c_malloc(sizeof(DATA));
  dat->ID           = (char**)c_malloc((NoInd*NoLoci)*(sizeof(char*)));
  dat->markers      = (char**)c_malloc((NoInd*NoLoci)*(sizeof(char*)));
  dat->all_pairs    = (int**)c_malloc((NoInd*NoLoci)*(sizeof(int*)));
  dat->rawall_pairs = (float**)c_malloc((NoInd*NoLoci)*(sizeof(float*)));

  for (i=0; i < NoInd*NoLoci; i++) {
    dat->ID[i]        = (char*)c_malloc((MAXCHAR)*(sizeof(char)));
    dat->markers[i]   = (char*)c_malloc((MAXCHAR)*(sizeof(char)));
    dat->all_pairs[i] = (int*)c_malloc((2)*(sizeof(int)));
    dat->rawall_pairs[i] = (float*)c_malloc((2)*(sizeof(float)));
    for (j=0; j < 2; j++) {
      dat->all_pairs[i][j]    = 0;
      dat->rawall_pairs[i][j] = 0;
    }
  }
  return(dat);
}

/* 
   function to examine if datastr contains a string to indicate
   missing data
*/
int missingdata(char *datastr, char *miss[], int miss_sym) {
  int i;

  for (i=0; i < miss_sym; i++) {
    if (strcmp(datastr, miss[i]) == 0) return 1;
  }
  return 0;
}


/* read in an .abi file */
int marker_data(char *ped_file) {
  FILE *ifp;
  int i,j,len,miss_hit=0;
  char orig[MAXCHAR]="";
  char c;
  char *miss[3]={"Failed","failed","-"};/* missing symbols */
  int  miss_sym=3;
  int headerline = 1;
  int nfields = 0;

  printf("The missing data indicators are\n");
  for (i=0;i<miss_sym;i++) {
    printf("%s ",miss[i]);
  }
  printf("\n");

  i=0;

  if ((ifp=fopen(ped_file,"r")) == NULL) {
    printf("Can't open file %s\n",ped_file);
    exit (1);
  }

  while (!feof(ifp)) {

    if (getstring(orig,ifp) == -1) { /* EOF*/
      break;
    }	
    strcpy(dat->ID[i],orig);
    
    getstring(orig,ifp);
    strcpy(dat->markers[i],orig);

    nfields= 0;
    /* read the binned integer allele lengths */
    for (j = 0; j< 2; j++) {
      getstring(orig,ifp);
      if ( missingdata(orig, miss, miss_sym) == 0) {
	dat->all_pairs[i][j]=atoi(orig);
	nfields += sscanf(orig,"%d",&dat->rawall_pairs[i][j]);
      }
      else {
	dat->all_pairs[i][j]=0;
	nfields++;
      }
    }
    
    /* read the raw decimal allele lengths */
    for (j = 0; j< 2; j++) {
      getstring(orig,ifp);
      if ( missingdata(orig, miss, miss_sym) == 0) {
	/* dat->rawall_pairs[i][j]=atof(orig); */
	nfields += sscanf(orig,"%f",&dat->rawall_pairs[i][j]);
	}
      else {
	dat->rawall_pairs[i][j]=0;
	nfields++;
      }
    }
    if (headerline && nfields != 4) {
      /* all fields pm the first line were not sucessfully converted 
       * so we have a headerline */
      headerline = 0;
    }
    else {
      i++;
    }
  }
  fclose(ifp);

  if (op_.Debug == 11) {
    for (j = 0; j < 5; j++) {
      printf("ID[%d]=%s,markers[%d]=%s,all_pairs=(%d,%d),rawall_pairs=(%f,%f)\n",
	     j,dat->ID[j],j,dat->markers[j],
	     dat->all_pairs[j][0],dat->all_pairs[j][1],
	     dat->rawall_pairs[j][0],dat->rawall_pairs[j][1]);
    }
  }

  return(i);
}

/* read in the .ped file to merge with the .ABI file */
int read_in_partped(char *template_file,int abi_sample) {
  int i,j,k;
  FILE *ifp;
  char orig[MAXCHAR]="";
  char temp;
  int temp_length,len=1;
  int headerline = 1;
  int nfields = 0;

  i=0;
  if ((ifp=fopen(template_file,"r")) == NULL) {
    fprintf(stderr,"Can't open input file %s\n",template_file);
    exit(1);
  }

  if (abi_sample == 1) {
    temp_length=7;
  }
  else{
    temp_length=6;
  }
  /* First read header line 
  
  for (j=0; ! op_.idxoutput && j < temp_length; j++) {    
    len=getstring(orig,ifp);
    if (op_.Debug == 11) {
      printf("Header line=%s\n",orig);
    }
  }
  */

  while (len != -1) { /* len is also a flag to indicate EOF*/
    len=getstring(orig,ifp);
    if (len == -1) {
      break;
    }
    if (i >= MAXIND) {
      fprintf(stderr,"WARNING - Maximum number of individuals exceeds limit of %d\n",i);
      fprintf(stderr,"Either reduce no of individuals or modify MAXIND and recompile\n");
      fprintf(stderr,"Exiting...\n");
      exit(1);
    }
    
    strcpy(ped->pedids[i],orig);
    getstring(ped->indids[i],ifp);
    getstring(ped->faids[i],ifp);
    getstring(ped->maids[i],ifp);

    nfields = 0;
    for (j = 0; j < liab; j++) {
      getstring(orig,ifp);
      nfields += sscanf(orig,"%d",&ped->misc[i][j]);
    }

    if (abi_sample == 1) {
      len=getstring(orig,ifp);
      strcpy(sample_id[i],orig);
    }
    else{
      /*  we no longer need to read in blank space to the new line
	  getstring(orig,ifp);  read any blank space before the newline 
      */
    }
    
    if (headerline && nfields != liab) {
      headerline = 0;
      if (op_.Debug == 11) {
	printf("Header line found: %s %s %s %s ... %s\n",
	       ped->pedids[i],ped->indids[i],ped->faids[i],ped->maids[i],orig);
      }
    }
    else {

      if (op_.Debug == 11 && i < 10) {
	printf("here i=%d,%s %s %s %s",
	       i,ped->pedids[i],ped->indids[i],ped->faids[i],ped->maids[i]);
	if (abi_sample == 1) printf(" %s", sample_id[i]);
	printf("\n");
      }

      i++;

    }
  }
  return(i);
}

/* write pedigree - putting the right markers next to the right IDs*/
void create_pre(int TOT,int abi_sample) {
  int i,j,k,old_marker=0;
  int found=0;
  char temp1[MAXCHAR];

  j=0;
  for (i=0; i < TOT; i++) { /* length of ped file */
    found=0;
    for (k=0; k < IND; k++) { /* going through the template file */

      if (abi_sample == 1) {
	strcpy(temp1,sample_id[k]);
	if (op_.Debug == 13) {
	  printf("sample_id to be checked=%s\n",sample_id[k]);
	}
      }
      else{
	strcpy(temp1,ped->indids[k]);
      }

      if (strcmp(dat->ID[i],temp1) == 0) {
	/* to speed things up a bit, usually all markers are together */
	if (op_.Debug == 11) {
	  printf("first marker =%s,first marker in abi file=%s\n",
		 mark_ord[old_marker],dat->markers[i]);
	}

	if (strcmp(mark_ord[old_marker],dat->markers[i])==0) {
	  ped->m[k][j][0]=dat->all_pairs[i][0];
	  ped->m[k][j][1]=dat->all_pairs[i][1];
	  ped->rawm[k][j][0]=dat->rawall_pairs[i][0];
	  ped->rawm[k][j][1]=dat->rawall_pairs[i][1];

	  if (op_.Debug == 13) {
	    printf("marker alleles for ind %s,Marker %d=(%d,%d)raw (%f,%f)\n",
		   ped->indids[k],j,ped->m[k][j][0],ped->m[k][j][1],
		   ped->rawm[k][j][0],ped->rawm[k][j][1]);
	  }
	}
	else{

	  for (j=0;j<L;j++) { /*going through the markers */
	    if (strcmp(mark_ord[j],dat->markers[i])==0) {
	      ped->m[k][j][0]=dat->all_pairs[i][0];
	      ped->m[k][j][1]=dat->all_pairs[i][1];
	      ped->rawm[k][j][0]=dat->rawall_pairs[i][0];
	      ped->rawm[k][j][1]=dat->rawall_pairs[i][1];
	      old_marker=j;
	      break;
	    }
	  }

	}
	found=1;
      }
      if (found==1) {
	break;
      }
    }
    if (found==0) {
      printf("Can't find data to go with line %d in ABI file\n",i+1);
      printf("ID=%s,Marker=%s,Allele1=%d,Allele2=%d\n",dat->ID[i],dat->markers[i],dat->all_pairs[i][0],dat->all_pairs[i][1]);
    }
  }				
}

void rawabi(char *mark_file,char *template_file,char *ped_file,int abi_sample) {
  int i,len,TOT=0;
  FILE *ifp;
  char orig[MAXCHAR];

  /* read the marker file and store the markers in order in mark_ord array
   * Store the number of markers found in L
   * Allocate data,
   * read the template file to obtain pedigree info for each person
   * read in the ABI file which contains the raw and rounded allele 
   * lengths found at each marker (marker_data)
   * Finally create a pedigree file 
   */


  if ((ifp=fopen(mark_file,"r")) == NULL) {
    printf("Can't open file %s\n",mark_file);
    exit(1);
  }


  i=0;
  while ( getstring(orig,ifp) != -1) { /* -1 is EOF */
    if (strlen(orig) > 0 && orig[0] != '#') {    
      /* only process lines containing text */
      strcpy(mark_ord[i],orig);
      i++;
    }
  }
  fclose(ifp);
  

  L=i;          /* Set L to be number of marker lines */
  if (L >= MAXLOCI) {
    fprintf(stderr,
	    "WARNING - Maximum number of markers exceeds limit of %d\n",L);
    fprintf(stderr,
	    "Either reduce no of markers or modify MAXLOCI and recompile\n");
    fprintf(stderr,"Exiting...\n");
    exit(1);
  }

  if (op_.Debug == 11) {
    for (i=0; i < L; i++) {
      printf("Marker %d=%s\n",i,mark_ord[i]);
    }
  }

  IND=read_in_partped(template_file,abi_sample);	
  if (op_.Debug == 11) printf("IND=%d\n",IND);

  dat=allocate_data(IND,L);

  TOT=marker_data(ped_file);
  if (op_.Debug == 11) printf("TOT=%d\n",TOT);

  create_pre(TOT,abi_sample);
  if (op_.Debug == 11 || op_.Debug == 13) print_ped();
  
}
