#include "raw.h"

/* 
   last change 18/4/2001 
   reads in raw ABI files 

   $Id: rawlink.c,v 2.22 2002/10/14 05:39:46 bahlo Exp $

   $Log: rawlink.c,v $
   Revision 2.22  2002/10/14 05:39:46  bahlo

   Minor bug in -ef option fixed.

   Revision 2.21  2002/10/11 07:24:01  bahlo

   This is a minor change regarding the -ef option. The minimum no of
   alleles is now 10 with an equal allele frequency of 0.1. Markers that
   have a
   larger number of allleles will have 1/no of alleles as their equal allele
   frequency. This was done because potentially alleles had too small a
   frequency producing low lod scores.

   Revision 2.20  2002/10/10 03:19:22  wilkinson
   Fixed an previously introduced error for inserting male penetrances
   in the .dat file.  The code was right, I misread the linkage manual
   and made it wrong. Now I'm making it right again

   Revision 2.19  2002/09/23 00:24:50  wilkinson
   fixed potential bug (for other programs) due to no space between
   number_of_markers and << field seperator in.pre files.

   Revision 2.18  2002/09/10 05:41:31  wilkinson
   if one allele is unbinned- both alleles set to zero

   Revision 2.17  2002/09/10 05:29:40  wilkinson
   added a check that all non-zero alleles are rebinned

   Revision 2.16  2002/09/06 07:16:28  wilkinson
   updated code to allow for incorporation of existing allele frequencies

   Revision 2.15  2002/09/06 03:59:02  wilkinson
   updated rebinning function to handle case where we have a small family and
   we wish to use rebinned alleles from a seperate study, where there are more
   alleles at a given marker than is in the family. The next step is to use
   the allele freq info from the larger study in a .dat file.

   Revision 2.14  2002/09/05 00:36:19  wilkinson
   modified ouput .dat file so that distances less than 0.01 are set to 0.000001
   this is because genehunter will crash if it comes across a distance of 0
   between markers.

   Revision 2.13  2002/07/18 01:34:03  wilkinson
   modified reading of marker and distance (recomb frac) files to ignore
   lines beginning with a '#'. When reading distances, it will store the
   distance after the # and add that to the the distance to the next marker.

   Revision 2.12  2001/12/10 01:19:03  wilkinson
   added an option to produce marker files in .idx format for use in
   McPeek and Sun's prest program for detecting misspecified relationships
   (AmJHG 66:1076-1094)

   Revision 2.11  2001/11/22 03:01:38  wilkinson
   fixed bug in printing of allele frequecies in .dat files
   The format specfication was changed from 1.3f to 1.3g to
   handle zero alleles which we reset to a min value of 1.0e-6.
   the f format meant we were printing out 0.000, however we
   are now printing out 1e-6.

   Revision 2.10  2001/11/16 00:27:16  wilkinson
   Enabled the use of an input file in placed of command line arguments.
   The file template file inputs lists the general style required.

   Revision 2.9  2001/11/14 04:33:46  wilkinson
   add warnings if unable to open files for reading or writing

   Revision 2.8  2001/11/09 01:30:24  wilkinson
   added a debug flag on the command line
   added a check to ensure that zeroed frequencies (ie present in affected or
   sibs but not in founders) are set to a value of MINFREQUENCY

   Revision 2.7  2001/11/08 06:09:20  bahlo

   changed frequencies function to accomodate -ef option

   Revision 2.6  2001/11/02 00:25:41  wilkinson
   Fixed bug in wirting out of .dat files (we were writing out N-1 Loci
   instead of N (I wasn't taking into account affection locus). updated
   documentation


   Revision 2.0  2001/10/08 02:31:33  wilkinson
   Fixed bug with reading in of marker files (reading in one less than
   the desired number).
   Made getline function more robust to allow multiple spaces or tabs to
   be handled (originally all field could only be seperated by a single space
   or tab).  getline also now handles whitespace after the field being read
   Program also handles extra blank lines at the end of input files


*/




/* checks whether the distances are recomb fracts. If not changes them to 
fracs from cm via Kosambi's map function*/
void cm_to_fract() {
  int i,j;
  double temp;
  for (i=0; i < L; i++) {
    if (theta[i] > 0.5) { /* test whether ALL distances are in recomb fractions ie <0.5 */
      for (j=0; j < L; j++) {
	temp=exp(4.0*theta[j]/100.0)-1.0;
	temp/=(exp(4.0*theta[j]/100.0)+1.0);
	temp*=0.5;
	theta[j]=temp;
      }
      break;
    }
  }
}

double to_recombfrac(double cm) {
  double recombfrac;
  
  recombfrac = 0.5 * (exp(4*cm/100.0) - 1.0) / (exp(4*cm/100.0) + 1.0);
  return recombfrac;
}


/* recoding of alleles: writing out a new PEDIGREE*/

/* check function: checks whether the allele is novel to the current array 
of allele labels*/

int check_all(int allele,int locus,int a[L][MAXALL]) {
  int i;
  for (i=0;i < NoAlleles[locus]; i++) {
    if (allele == a[locus][i])
      return(0);
  }
  return(1);	
}

void check_a_matrix(int a[L][MAXALL]) {
  int i,j=0;
  for (i=0; i <L; i++) {
    for (j=0; j < NoAlleles[i]; j++) {
      printf(" a[%d][%d]=%d",i,j,a[i][j]);
    }	
    printf("\n");
  }
}

/* 
   collect the different alleles at each loci into array a with indicies
   a[marker j][kth unique allelle]
   NoAlleles[marker j] stores the number of unique allele for each marker
*/ 

void collect(int a[L][MAXALL]) {
  int i,j,k,ch,l;

  for (j=0; j < L; j++) {

    a[j][0] = ped->m[0][j][0];

    NoAlleles[j]=1;

    if (ped->m[0][j][1] !=a [j][0]) {
      NoAlleles[j]++;
      a[j][1]=ped->m[0][j][1];
    }

    for (i=1; i < IND;i++) {
      for (k=0; k < NoAlleles[j]; k++) {
	for (l=0;l < 2; l++) {
	  ch=check_all(ped->m[i][j][l],j,a);
	  /*printf("m[%d][%d][%d]=%d",i,j,l,ped->m[i][j][l]);*/
	  if (ch == 1) {
	    a[j][NoAlleles[j]]=ped->m[i][j][l];
	    NoAlleles[j]++;
	  }
	}
      }
    }
  }
}

/* reordering array of alleles */
void reorder(int a[L][MAXALL]) {
  int j,i,pass,hold; /* I nicked this from p.225 of Deitel & Deitel */
  for (j=0; j < L; j++) {
    for (pass=1; pass < NoAlleles[j]; pass++) {
      for (i=0; i < NoAlleles[j]-1; i++) {
	if (a[j][i] > a[j][i+1]) {
	  hold=a[j][i];
	  a[j][i]=a[j][i+1];
	  a[j][i+1]=hold;
	}
      }
    }
  }
}

/* remove 0 alleles */
void zero(int a[L][MAXALL]) {
  int marker_j,k;

  for (marker_j = 0; marker_j < L; marker_j++) {
    if (a[marker_j][0] == 0) {
      for (k = 0; k < NoAlleles[marker_j] - 1; k++) {
	a[marker_j][k] = a[marker_j][k+1];
      }
      NoAlleles[marker_j]--;

      if (NoAlleles[marker_j] <= 0)         /* marker is missing */
	HandleMissingMarker(marker_j);
	
    }
  }
}

/* handle the case where we have a missing marker.
 * Default behaviour is to issue a warning message to the user.
 */
void HandleMissingMarker(int marker_j) {
  int i;

  fprintf(stderr,"\n");
  for (i = 0; i < 70; i++) 
    fprintf(stderr,"#"); 
  fprintf(stderr,"\n\n");
  fprintf(stderr,
	  "WARNING - No data exists on Marker %s specified in marker list (%d of %d)\n",
	  mark_ord[marker_j],marker_j+1,L);
  if (op_.ignoreMissingMarkers) {
    fprintf(stderr,
      "Program will continue, and ignore this marker in the output .pre and .dat files\n\n");
    for (i = 0; i < 70; i++) 
      fprintf(stderr,"#"); 
    fprintf(stderr,"\n\n");
  }
  else {
    fprintf(stderr,"Please edit the marker list file and the map distance file to\n");
    fprintf(stderr,"remove this marker, and then rerun the program.\n");
    fprintf(stderr,"Exiting....\n\n");
    exit(1);
  }
}


/* 
 * Read in a file containing rebinned allele values.
 * bin1[i][j][0] stores old allele value
 * bin1[i][j][1] stores new allele value
 * The file format is a
 * Marker1
 * Old Allele1  New Allele1
 * Old Allele2  New Allele2
 * 
 * Blank lines seperate the markers. 
 * Markers are numeric from 1 .. Total No  of Markers (L)
 *
 * Our approach is to read each line
 * if blank line found, reset the marker_counter;
 * if marker line found, set i to (marker no - 1), reset j to 0
 * if OriginalAllele = TRUE, store original allele value, 
 * if OriginalAllele = FALSE, store new allele value and increment j
 * Original Allele switches value each time we store data in bin1 array
*/
void newbins(int bin1[L][MAXALL][2],char *rebinfile, int a[L][MAXALL]) {
  int i,j,OriginalAllele,temp,MarkerCounter,len,rc=1;
  char DataField[MAXCHAR],str1[]="\n";
  FILE *ifp;

  if ((ifp=fopen(rebinfile,"r")) == NULL) {
    printf("Can't open rebinfile %s\n",rebinfile);
    exit(1);
  }

  for (i=0; i < L; i++) {
    for (j=0; j < MAXALL; j++) {
      bin1[i][j][0]=0;
      bin1[i][j][1]=0;
    }
  }
  i=0;
  j=0;
  MarkerCounter = 0;
  OriginalAllele = TRUE;

  while ( (rc=getstring(DataField,ifp)) != -1) { /* EOF tester */
       
    if (strlen(DataField) == 0) {
      MarkerCounter = 0;
    }
    else {
      if (MarkerCounter == 0) { 
	i = atoi(DataField) - 1;
	j = 0;
      }
      else {
	if (OriginalAllele == TRUE) {
	  bin1[i][j][0]  = atoi(DataField);
	  OriginalAllele = FALSE;
	}
	else {
	  bin1[i][j][1]  = atoi(DataField);
	  OriginalAllele = TRUE;
	  j++;
	}
      }
      MarkerCounter++;
    }
  }

  i++;
  if (i != L) {
    printf("Not enough markers in .rebin file\n");
    printf("i=%d,L=%d\n",i,L);
    exit(1);
  }
  /* now check that the number of alleles in the data is not greater than 
     the number of alleles in the rebin file.  In that case we won't be 
     able to rebin some alleles (they would retain their old values)
  */

  for (i = 0; i < L; i++) {
    for (j=0; j < MAXALL && bin1[i][j][0] > 0; j++) {
      a[i][j]=bin1[i][j][1];
    }
    if (NoAlleles[i] > j) {
      printf("Not enough alleles for marker %d in .rebin file\n",i);
      printf("%d Alleles in data but only %d alleles in rebin file\n",NoAlleles[i],j);
      exit(1);
    }
    /* set the number of alleles to the number of rebinned alleles */
    NoAlleles[i] = j;
  }

}


/* In some circumstances the number of alleles observed in the 
   pedigree may be less than the number used inthe rebinning process
   eg when we want to use alleles from a large dataset (which has
   good freq info) in a small family.  we need to rebin the alleles
   to the codes of the larger study
   So I've modified the code to use the number of binned alleles
   rather than the NoAlleles[i] seen in the sample.
   we should probably raise an error if NoAlleles[i] is 
   greater than the number of rebinned alleles
    */

void print_bin(int bin1[L][MAXALL][2]) {
  int i,j,cnt=0;
  printf("The rebinned alleles are:\n");
  for (i=0; i < L; i++) {
    cnt=0;
    
    /* for (j=0; j < MAXALL && bin1[i][j][0] > 0; j++) { */

    for (j=0; j < NoAlleles[i]; j++) {
      if (bin1[i][j][0] > 0) {
	printf("%d->%d\n",bin1[i][j][0],bin1[i][j][1]);
	cnt++;
      } 
    }
    if (cnt > 0) printf("\n");
  }
}

void rebinned(int a[L][MAXALL],int bin1[L][MAXALL][2]) {
  int i,j,k,l;
  int not_rebinned;

  for (i=0; i < IND; i++) {
    for (j=0; j < L; j++) {
      for (l=0; l < 2; l++) {
	not_rebinned = 1;
	/*	for (k=0; k < MAXALL && bin1[j][k][0] > 0; k++) { */

	for (k=0; k < NoAlleles[j]; k++) { 
	  if (ped->m[i][j][l] != 0 && bin1[j][k][0] != 0) {
	    if (ped->m[i][j][l] == bin1[j][k][0]) {
	      ped->m[i][j][l] = bin1[j][k][1];
	      not_rebinned = 0;
	      break;
	    }
	  }
	}
	if (not_rebinned && ped->m[i][j][l] != 0) {
	  if (op_.ignoreMissingMarkers) {
	    printf("WARNING: REBINNING ERROR: ID %3d marker %2d: No rebin value for Allele %3d. Setting to zero\n",
		   i,j,ped->m[i][j][l]);
	    ped->m[i][j][0] = 0;
	    ped->m[i][j][1] = 0;
	    break;
	  }
	  else {
	    printf("WARNING: REBINNING ERROR: ID %3d marker %2d: No rebin value for Allele %3d. Exiting...\n",
		   i,j,ped->m[i][j][l]);
	    exit(1);
	  }
	}

      }
    }
  }
}

/* writing out a new PEDIGREE with alleles relabelled from 
   1..number of unique alleles (up to NoAlleles[marker j]) at each, 
   allele relabellings will be written out to a new file and a 
   new pedigree file with relabelled alleles is also created 
*/
void new_ped(int a[L][MAXALL]) {
  int i,j,k,l;
  for (i=0; i < IND; i++) {
    for (j=0; j < L; j++) {
      for (l=0; l < 2; l++) {
	for (k=1; k <= NoAlleles[j]; k++) {
	  if (ped->m[i][j][l] != 0) {
	    if (ped->m[i][j][l] == a[j][k-1]) {
	      ped->m[i][j][l]=k;
	      break;
	    }
	  }
	}
      }
    }
  }
}

/* 
   calculation of allele frequencies for each marker based upon 
   unaffected founder population
   if existing allele freqs are used format should be
   allele_value frequency
   note if you are rebining the alele_values should be the raw allele values
   not the binned (ie 1..N) values.

*/
void frequencies(int a[L][MAXALL],int relabel) {
  int i,j,l,k;
  int cnt[L];
  int MarkerCounter = 0;
  int  alleleColumn = TRUE;
  int  rc,n=0;
  char DataField[MAXCHAR];
  FILE *ifp,*ofp;

  for (j=0; j < L; j++)
    cnt[j]=0;

  if (op_.useExistingFreqs) {
    if ( op_.freqFile == NULL || 
	 (ifp=fopen(op_.freqFile, "r")) == NULL) {
      printf("Can't open file %s\n",op_.freqFile);
      exit(1);
    }
    /* print out freq info for debugging purposes
       if ((ofp=fopen("freqs.out", "w")) == NULL) {
       printf("Can't open file freqs.out\n");
       exit(1);
       }
    */
  }
  else if (op_.freqFile != NULL) {
    if ((ofp=fopen(op_.freqFile, "w")) == NULL) {
      printf("Can't open file %s\n",op_.freqFile);
      exit(1);
    }
  }
  else {
    if ((ofp=fopen("freqs.out", "w")) == NULL) {
      printf("Can't open file freqs.out\n");
      exit(1);
    }
  }

  if (op_.useExistingFreqs) {
    MarkerCounter = 0;
    alleleColumn = TRUE;
    n=0;
    while ( (rc=getstring(DataField,ifp)) != -1) { /* EOF tester */
       
      if (strlen(DataField) == 0) {
	MarkerCounter++;
	n = 0;
      }
      else {
	if (alleleColumn == TRUE) {
	  alleleColumn = FALSE;
	}
	else {
	  sscanf(DataField,"%lf",&freq[MarkerCounter][n]); 
      
	  if (freq[MarkerCounter][n] < MINFREQUENCY) 
	    freq[MarkerCounter][n] = MINFREQUENCY;
	  alleleColumn = TRUE;
	  n++;
	}
      }
    }
    close(ifp);
  }
  else {

    if(op_.allelefreqs==FALSE){
      for (i=0; i < IND; i++) {		/* loop over the individuals */
	
	if ( (strcmp(ped->faids[i],ped->maids[i]) != 0) || 
	     (ped->misc[i][1] == AFFECTED))
	  continue;
	/* ok - we have an unaffected founder
	 * now loop over the L markers, and over the two alleles
	 */
	for (j=0; j < L; j++) {	
	  for (l=0; l < 2; l++) {	
	    /* make sure we don't double count males when dealing with 
	     * the x-chromosome
	     */
	    if (l == 1 && op_.sexlinked && ped->misc[i][0] == MALE)
	      continue;
	    /* now hunt for the current allele. When we find it, 
	     * increment a counter for this allele (freq), and for the
	     * number of indivividuals for this marker (cnt)
	     */
	    for (k=1; k <= NoAlleles[j]; k++) {
	      if (ped->m[i][j][l] != 0) {
		if ( (relabel == 1 && ped->m[i][j][l] == k) ||
		     (relabel != 1 && ped->m[i][j][l] == a[j][k-1]) ) {
		  freq[j][k-1]+=1.0;
		  cnt[j]++;
		  break;
		}
	      }
	    }
	  }
	}
      }
    }	
    else{
	/* changed by MB: in small pedigrees the equal allele freqs based
	solely on NoAlleles leads to large frequencies. As an alternative
	for microsats use 10 alleles with equal freq of 0.1, unless the
	observed no of alleles is greater in the pedigree, in which case 
	use NO_Alleles. */
     	for (i=0; i < IND; i++) {		/* loop over the individuals */
			for (j=0; j < L; j++) {	
				if(NoAlleles[j]>10){
					for (k=1; k <= NoAlleles[j]; k++) {
	    				freq[j][k-1]=1.0/(double)(NoAlleles[j]);
	  				}
				}
				else{
					for (k=1; k <= 10; k++) {
	    				freq[j][k-1]=0.1;
	  				}
				}
			}
      	}
    }
    /* close for i =0..IND loop */
    for (j=0; j < L; j++) {
      for (k=0; k < NoAlleles[j]; k++) { 
	
	if(op_.allelefreqs == FALSE){
	  freq[j][k]/=(double)(cnt[j]);
	  if (freq[j][k] < MINFREQUENCY) freq[j][k] = MINFREQUENCY;
	}
      }
    }
  }

  for (j=0; j < L; j++) {
    for (k=0; k < NoAlleles[j]; k++) { 
      
      if (relabel == 1) {
	if(op_.allelefreqs){
	  printf("Allele %d->%d, freq[%d][%d]=%lf\n",a[j][k],k+1,j,k,freq[j][k],freq[j][k]);
	  fprintf(ofp,"Allele %d->%d, freq[%d][%d]=%lf\n",a[j][k],k+1,j,k,freq[j][k]);
	}
	else{
	  printf("Allele %d->%d, freq[%d][%d]=%lf (%2.0lf of %2d)\n",
		 a[j][k],k+1,j,k,freq[j][k],freq[j][k]*(double)(cnt[j]),cnt[j]);
	  fprintf(ofp,"Allele %d->%d, freq[%d][%d]=%lf (%2.0lf of %2d)\n",
		  a[j][k],k+1,j,k,freq[j][k],freq[j][k]*(double)(cnt[j]),cnt[j]);
	}
      }
      else{

	if(op_.allelefreqs){
	  printf("Allele %d, freq[%d][%d]=%lf\n",a[j][k],j,k,freq[j][k]);
	  fprintf(ofp,"Allele %d, freq[%d][%d]=%lf\n",a[j][k],j,k,freq[j][k]);
	}
	else{
	  printf("Allele %d, freq[%d][%d]=%lf (%2.0lf of %2d)\n",
		 a[j][k],j,k,freq[j][k],freq[j][k]*(double)(cnt[j]),cnt[j]);
	  fprintf(ofp,"Allele %d, freq[%d][%d]=%lf (%2.0lf of %2d)\n",
		  a[j][k],j,k,freq[j][k],freq[j][k]*(double)(cnt[j]),cnt[j]);
	}

      }
      if (relabel == 1) {a[j][k]=k+1;}		
    }
  }
}

/* writes a new .dat file */
void new_dat(char *recfile,char *datfile,int a[L][MAXALL],int relabel) {
  int this,l,j,k,b=0,NoMissingMarkers=0;
  double temptheta,temp,distance;

  FILE *ofp,*ifp;	
  char DataField[MAXCHAR];


  if ((ofp=fopen(datfile,"w")) == NULL) {
    printf("Can't open file %s\n", datfile);
    exit(1);
  }
  if ((ifp=fopen(recfile,"r")) == NULL) {
    printf("Can't open file %s\n", recfile);
    exit(1);
  }

  /* Read in distances from recfile.  Don't record
   * any lines that start with a #, but do record their
   * values, since we need to interpolate distances
   * over this missing marker(s). 
   * Note there could be several adjacent markers excluded
   * so we handle this case by using temp and temptheta
   */
 
  j=0;
  temptheta=temp=0.0;
  while ( getstring(DataField, ifp) != -1) {
    if (strlen(DataField) == 0)  { continue; }
    else if (DataField[0] == '#') {
      sscanf(DataField+1,"%lf",&temp);
      temptheta+=temp;
      continue;
    }
    else {
      sscanf(DataField,"%lf",&theta[j]);
      theta[j]+=temptheta;
      temptheta=0.0;
    }

    if (op_.Debug == 11) {
      printf("theta[%d]=%lf\n",j,theta[j]);
    }

    if (NoAlleles[j] == 0) 
      NoMissingMarkers++;
    j++;
  }
  if (j != L) {
    fprintf(stderr, 
	    "WARNING - Insufficent number of distances listed in %s\n",
	    recfile);
    exit(1);
  }
  /*
  for (j=0; j < L; j++) {


    fscanf(ifp,"%lf",&theta[j]);
    if (op_.Debug == 11) {
      printf("theta[%d]=%lf\n",j,theta[j]);
    }

    if (NoAlleles[j] == 0) 
      NoMissingMarkers++;
  }
  */

  /* the number of Loci is the number of present markers + 1 as affection 
   * status is another loci */
  fprintf(ofp,"%d 0 %d 5 << NO. OF LOCI, RISK LOCUS, SEXLINKED (IF 1) PROGRAM\n", 
	  L + 1 - NoMissingMarkers, op_.sexlinked);
  fprintf(ofp,"0 0.0 0.0 0 << MUT LOCUS, MUT RATE, HAPLOTYPE FREQUENCIES (IF 1)\n");
  for (l=0;l < L + 1 - NoMissingMarkers; l++) {fprintf(ofp,"%d ",l+1);}
  fprintf(ofp,"\n");
  fprintf(ofp,"1 2 << AFFECTATION, NO. OF ALLELES\n");
  fprintf(ofp,"0.88 0.12 << GENE FREQUENCIES (change if required)\n");
  fprintf(ofp,"1 << NO. OF LIABILITY CLASSES (change if required)\n");
  fprintf(ofp,"0.003 0.900 0.900 << PENETRANCES (change if required\n");
  if (op_.sexlinked)
    fprintf(ofp,"0.0 0.5           << MALE PENETRANCES (change if required) \n");

  NoMissingMarkers = 0;
  for (l=0; l < L; l++) {

    if (NoAlleles[l] == 0) {
      NoMissingMarkers++;
      continue;
    }
    if (relabel == 0 || op_.allelefreqs) {
		if(relabel == 0){
      		fprintf(ofp, "3  %d << ALLELE NUMBERS, NO. OF ALLELES (Marker %d)\n",
	     	 a[l][NoAlleles[l]-1],l+1 - NoMissingMarkers);
		}
		else{ /* equal allele freqs */
      	fprintf(ofp, "3  10 << ALLELE NUMBERS, NO. OF ALLELES (Marker %d)\n",l+1 - NoMissingMarkers);
		}
    }
    else{
      fprintf(ofp, "3  %d << ALLELE NUMBERS, NO. OF ALLELES (Marker %d)\n",
	      NoAlleles[l],l+1 - NoMissingMarkers);
    }
    if (op_.Debug == 6) {
      printf("max allele is %d\n",a[l][NoAlleles[l]-1]);
    }
	/* printing out the actual frequencies.
	for equal allele frequencies (currently limited to 10 alleles
	just tack on the missing allele freqs after the max allele freq*/
    for (k=1;k <= a[l][NoAlleles[l]-1]; k++) { 
      /* a[l][NoAlleles[l]-1] is the maximum sized allele*/
      this=0;
      for (j=0; j < NoAlleles[l]; j++) {
	if (k == a[l][j]) {
	  fprintf(ofp,"%1.3g ",freq[l][j]);
	  this=1;
	  break;
	}
      }
      if (this == 0) {
	fprintf(ofp, "0.0 ");
      }			
    }
	for(k=0;k<(10-NoAlleles[l]);k++){
		fprintf(ofp,"0.1 "); /*change if required or make into a subarg of ef*/
	}
    fprintf(ofp, "\n");
  }
  fprintf(ofp,"0 0 << SEX DIFFERENCE, INTERFERENCE (IF 1 OR 2)\n");

  /* modified by CRW 28Sept 01. The distance file format now includes
     a zero as the first entry (so the marker and distance files are 
     both of length L).  This eliminates the need to append a zero.
     fprintf(ofp, "0.0 ");
     for (b=0;b < (L-1);b++) {
  */ 

  for (temptheta = 0, l = 0; l < L; l++) {
    if (NoAlleles[l] == 0) {
      temptheta += theta[l];
    } else {
      distance = theta[l] + temptheta;
      if (distance > 0.01)
	fprintf(ofp, "%2.2lf ",distance);
      else
	fprintf(ofp, "0.000001 ");
      temptheta = 0;
    }
  }
  fprintf(ofp,"\n1 0.1 0.45 << REC VARIED, INCREMENT, FINISHING VALUE\n");
  fclose(ifp);
  fclose(ofp);

}

/* writes a new .idx file for use by McPeek and Sun's Prest program */
void new_idx(char *recfile,char *idxfile,int a[L][MAXALL],int relabel) {
  int this,l,j,k,b=0,NoMissingMarkers=0;
  double temptheta,temp;

  FILE *ofp,*ifp;	
  char DataField[MAXCHAR];

  if ((ofp=fopen(idxfile,"w")) == NULL) {
    printf("Can't open file %s\n", idxfile);
    exit(1);
  }
  if ((ifp=fopen(recfile,"r")) == NULL) {
    printf("Can't open file %s\n", recfile);
    exit(1);
  }

  /* Read in distances from recfile.  Don't record
   * any lines that start with a #, but do record their
   * values, since we need to interpolate distances
   * over this missing marker(s). 
   * Note there could be several adjacent markers excluded
   * so we handle this case by using temp and temptheta
   */
 
  j=0;
  temptheta=temp=0.0;
  while ( getstring(DataField, ifp) != -1) {
    if (strlen(DataField) == 0)  { continue; }
    else if (DataField[0] == '#') {
      sscanf(DataField+1,"%lf",&temp);
      temptheta+=temp;
      continue;
    }
    else {
      sscanf(DataField,"%lf",&theta[j]);
      theta[j]+=temptheta;
      temptheta=0.0;
    }
    if (op_.Debug == 11) {
      printf("theta[%d]=%lf\n",j,theta[j]);
    }

    if (NoAlleles[j] == 0) 
      NoMissingMarkers++;
    j++;
  }
  if (j != L) {
    fprintf(stderr, 
	    "WARNING - Insufficent number of distances listed in %s\n",
	    recfile);
    exit(1);
  }

  /* the number of Loci is the number of present markers 
   * note that affection status is not regarded as a loci here */
  fprintf(ofp,"%d << NO. OF LOCI\n", L - NoMissingMarkers);

  NoMissingMarkers = 0;
  for (l=0; l < L; l++) {

    if (NoAlleles[l] == 0) {
      NoMissingMarkers++;
      continue;
    }
    if (relabel == 0||op_.allelefreqs) {
		if (relabel == 0){
      	fprintf(ofp, "3  %d << ALLELE NUMBERS, NO. OF ALLELES (Marker %d)\n",
	      a[l][NoAlleles[l]-1],l+1 - NoMissingMarkers);
		 }
		 else{
      fprintf(ofp, "3 10 << ALLELE NUMBERS, NO. OF ALLELES (Marker %d)\n",l+1 - NoMissingMarkers);
		 }
    }
    else{
      fprintf(ofp, "3  %d << ALLELE NUMBERS, NO. OF ALLELES (Marker %d)\n",
	      NoAlleles[l],l+1 - NoMissingMarkers);
    }
    if (op_.Debug == 6) {
      printf("max allele is %d\n",a[l][NoAlleles[l]-1]);
    }
	/* the actual frequencies
	For equal allele freqs just tack on the missing no of alleles as equal freqs at the end
	*/
    for (k=1;k <= a[l][NoAlleles[l]-1]; k++) { 
      /* a[l][NoAlleles[l]-1] is the maximum sized allele*/
      this=0;
      for (j=0; j < NoAlleles[l]; j++) {
	if (k == a[l][j]) {
	  fprintf(ofp,"%1.3g ",freq[l][j]);
	  this=1;
	  break;
	}
      }
      if (this == 0) { /* this is the option for raw alleleles-lots of 0's in the file*/
	fprintf(ofp, "0.0 ");
      }			
    }
    fprintf(ofp, "\n");
  }

  /* McPeek and Sun's program requires map distances in recombination
   * fractions. They also require n-1 recomb fractions for n markers,
   * so I don't need my leading zero which is in the distance file
   */

  for (temptheta = 0, l = 1; l < L; l++) {
    if (NoAlleles[l] == 0) {
      temptheta += theta[l];
    } else {
      fprintf(ofp, "%5.3lf ", to_recombfrac(theta[l] + temptheta) );
      temptheta = 0;
    }
  }
  fprintf(ofp,"\n");
  fclose(ifp);
  fclose(ofp);
}


void rawlink(char *recfile, char *datfile, char *idxfile, int relabel, int rebin, char *rebinfile,int eqall) {
  int a[L][MAXALL];
  int bin1[L][MAXALL][2];
  int i,j;

  collect(a);

  if (op_.Debug == 5) {
    printf("after collect\n");
    check_a_matrix(a);
  }			

  reorder(a); /* 0 is now first member in array*/

  if (op_.Debug == 5) {
    printf("after reorder\n");
    check_a_matrix(a);
  }	

  zero(a);

  if (op_.Debug == 5) {
    printf("after zero\n");
    check_a_matrix(a);
  }	

  if (rebin == 1) {
    newbins(bin1,rebinfile,a);
    print_bin(bin1);
    rebinned(a,bin1);
    if (op_.Debug == 5) {
      printf("after rebin\n");
      check_a_matrix(a);
    }

    /* I have to go through all of this again because I need to fix up the order 
       I have moved the function of collect into newbins.
       This is because the new bin file specifies all the 
       new bins that can exist - so we no longer have to work them 
       out from  the data. As it stands, such a calculation may not
       be possible for a small pedigree that only contains a subset of the 
       new alleles.  Note we probably still want to reorder and zero, 
       incase there input bins were not in order

       collect(a);
       if (op_.Debug == 5) {
       printf("after collect in rebin\n");
       check_a_matrix(a);
       }			
    */

    reorder(a); /* 0 is now first member in array*/
    if (op_.Debug == 5) {
      printf("after reorder in rebin\n");
      check_a_matrix(a);
    }	
    zero(a);	
    if (op_.Debug == 5) {
      printf("after zero in rebin\n");
      check_a_matrix(a);
    }
  }

  if (relabel == 1) {
    new_ped(a);
  }

  frequencies(a,relabel);
  new_dat(recfile,datfile,a,relabel);

  if (op_.idxoutput) {
    new_idx(recfile,idxfile,a,relabel);
  }

  if (op_.Debug == 5) {
    printf("thetas prior to conversion\n");
    for (i=0; i <  L; i++)
      printf("theta[%d]=%lf\n",i,theta[i]);
  }

  /* cm_to_fract();*/

  if (op_.Debug == 5) {
    printf("thetas after conversion from cM to recomb fract\n");
    for (i=0; i < L; i++)
      printf("theta[%d]=%lf\n",i,theta[i]);
  }

}
